import sys
import networkx as nx
from carthagene.metamap import Groups
from carthagene.metamap.Report import node_name,node_ref
from rpy import *
from math import log10

has_yap=True
try:
    import yapgvb
except:
    has_yap=False

Annotations={'Unchecked':-1,'Valid':0,'Reordered':1,'Rearranged':2,'Broken':3}
colors={-1:'grey',0:'green',1:'blue',2:'orange',3:'red' }
xcoord={-1:1,0:1.2,1:1.4,2:1.6,3:1.8 }
class Bit():
    '''
    Common Interval between a metamap and an assembly
    '''
    def __init__(self,mklist,assembly):
        self.ibeg=mklist[0].code
        self.iend=mklist[-1].code
        self.mkbeg=mklist[0].name
        self.mkend=mklist[-1].name
        self.mk=mklist
        self.pos=[assembly.position[x.name][1] for x in mklist]
        self.bpbeg=min(self.pos)
        self.bpend=max(self.pos)
        self.assidx=[assembly.getmap().index(x.name) for x in mklist]
        self.inv=self.assidx[0]>self.assidx[-1]
    def __len__(self):
        return self.bpend-self.bpbeg
    def __str__(self):
        return '%d-%d %d %11s-%11s %d-%d %d-%d  %10s' % \
                    (self.ibeg,self.iend,len(self.mk),self.mkbeg,self.mkend,self.bpbeg,self.bpend,min(self.assidx),max(self.assidx),(self.inv and 'inverted') or '')

class Segment():
    '''
    CI between metamap and assembly composed of bits
    sorted according to their position on the metamap.
    It has a freely setable type attribute
    '''
    def __init__(self,type=None):
        self.bits=[]
        self.len=0
        self.ibeg=-1
        self.iend=-1
        self.bpend=-1
        self.pos=[]
        if type in Annotations.keys():
            self.type=type
        else:
            raise ValueError
    def add(self,bit):
        self.bits.append(bit)
        self.bits.sort(key=lambda x: x.ibeg)
        self.ibeg,self.mkbeg= (self.ibeg < 0 and (bit.ibeg,bit.mkbeg)) or \
            (bit.ibeg < self.ibeg and (bit.ibeg,bit.mkbeg)) or \
            (self.ibeg,self.mkbeg)
        self.iend,self.mkend= (self.ibeg < 0 and (bit.iend,bit.mkend)) or \
            (bit.iend > self.iend and (bit.iend,bit.mkend)) or \
            (self.iend,self.mkend)
        self.pos.append(bit.bpbeg)
        self.pos.sort()
        self.bpbeg=self.pos[0]
        self.bpend=max(max(bit.pos),self.bpend)
        self.mkbeg=self.bits[0].mk[0].name
        self.mkend=self.bits[-1].mk[-1].name
        self.len+=len(bit.mk)
    def printd(self,stream=sys.stdout,reversed=False):
        buf=''
        buf+='-- %s segment: %d markers spanning %d Kb from %d to %d\n' %\
            (self.type,self.len,(self.bpend-self.bpbeg)/1000,self.bpbeg,self.bpend)
        buf+='\t%-10s%10s %-24s %-20s %9s %-14s \n' % ('Map Index','# mk','Flanking Markers','Assembly segment','Size (Kb)','Assembly Index')
        for bit in self.bits:
            if reversed:
                bit.inv=not bit.inv
            buf+='\t %2d:%4d-%-4d %9d %11s-%11s %9d-%-9d  %9d  %4d-%-4d  %10s\n' % \
                (self.bits.index(bit),bit.ibeg,bit.iend,\
                     len(bit.mk),bit.mkbeg,bit.mkend,min(bit.pos),max(bit.pos),(max(bit.pos)-min(bit.pos))/1000,min(bit.assidx),max(bit.assidx),\
                     (bit.inv and 'inverted') or '')
        print >>stream,buf
    def __str__(self):
        return '[ %d - %d ] : size %d Kb (%d markers)' % \
            (self.bpbeg,self.bpend,(self.bpend-self.bpbeg)/1000,self.len)

class Track():
    def __init__(self,beg,end,type,seg=None):
        if type in Annotations.keys():
            self.beg=beg
            self.end=end
            self.type=type
            self.seg=seg
        else:
            raise ValueError
    def __str__(self):
        return 'beg %d end %d type %s seg %s' %(self.beg,self.end,self.type,self.seg)
        
class AnnotatedAssembly():
    def __init__(self,assembly,reversed=False):
        self.tracks=[]
        self.mkpos=[x[1] for x in assembly.position.values()]
        self.length=max(self.mkpos)
        self.bits=[]
        self.reversed=reversed
        self.sort()
    def add(self,seg):
        self.tracks.append(Track(seg.bpbeg,seg.bpend,seg.type,seg))
        for b in seg.bits:
            added=self.add_bit(b)
#                 if added is not b:
#                     print b,'is included in',added
        self.sort()
    def add_bit(self,bit):
        match=[b for b in self.bits if b.ibeg <= bit.ibeg and b.iend >= bit.iend]
        if len(match)==0:
            self.bits.append(bit)
            return bit
        else:
            return match[0]
    def sort(self):
        self.tracks.sort(key = lambda x: x.beg)
        self.bits.sort(key = lambda x:x.bpbeg)
    def __str__(self):
        tw=''
        for t in self.tracks:
            tw+='[%d - %d] %s\n'%(t.beg,t.end,t.type)
        return tw
    def plot(self,title='Asscheck.pdf',reversed=False,order=False):
        print 'plotting, order=',order
        ybeg=-self.tracks[0].beg
        yend=-self.tracks[-1].end
        if not order:
            r.pdf(file=title,width=5,height=-yend*0.2/1000000,pointsize=16)
        else:
            r.pdf(file=title,width=5,height=-yend*0.2,pointsize=16)
        r.par(mfrow=[1,2],mar=[0,1,0,1])
        ### COMPARING MAP AND ASSEMBLY
        r.plot_new()
        r.plot_window(ylim=[yend,ybeg],xlim=[0,1.5])
        prevbeg=0
        for t in sorted(self.tracks,cmp = lambda x,y: ((x.beg == y.beg) and ((y.end-y.beg) -(x.end-x.beg)) ) \
                            or x.beg-y.beg):
            if t.beg==prevbeg:
                continue
            for b in t.seg.bits:
                if (not reversed and not b.inv) or (reversed and b.inv) :
                    coul='wheat'
                else:
                    coul='grey'
                if len(b.mk)>1:
                    r.polygon([1,1,1.5,1.5],[-b.bpbeg,-b.bpend,-b.bpend,-b.bpbeg],col=coul,border='NA')
                else:
                    r.segments(1,-b.bpbeg,1.5,-b.bpbeg,col=coul)
            start=t.beg
            ## size of gaps between bits
            gapsize=((t.seg.bpend-t.seg.bpbeg)-sum([(x.bpend-x.bpbeg) for x in t.seg.bits]))/(len(t.seg.bits)-1)
            for b in sorted(t.seg.bits,key=lambda x:x.ibeg,reverse=reversed):
                beg=int(start)
                end=int(start+len(b))
                if (not reversed and not b.inv) or (reversed and b.inv) :
                    coul='wheat'
                else:
                    coul='grey'
                if len(b)>0:
                    r.polygon([0,0,0.5,0.5],[-beg,-end,-end,-beg],col=coul,border='NA')
                else:
                    r.segments(0,-beg,0.5,-beg,col=coul)
                r.segments(1,-b.bpbeg-0.5*len(b),0.5,-beg-0.5*len(b))
                start=end+gapsize
            prevbeg=t.beg
        ###ANNOTATED ASSEMBLY
        r.plot_new()
        r.plot_window(ylim=[yend,ybeg],xlim=[0,2])
        if not order:
            r.axis(2,at=[x for x in r.seq(0,yend,-1000000)],labels=False,col="gray")
            r.axis(2,at=[x for x in r.seq(0,yend,-5000000)],labels=[str(-x/1000000) for x in r.seq(0,yend,-5000000)])
        else:
            r.axis(2,at=[x for x in r.seq(0,yend,-1)],labels=False,col="gray")
            r.axis(2,at=[x for x in r.seq(0,yend,-5)],labels=[str(-x) for x in r.seq(0,yend,-5)])
        r.axis(4,at=[-x for x in self.mkpos],labels=False,col='gray')
        r.title(ylab='Position (Mb)')
        for t in self.tracks:
            coul=colors[Annotations[t.type]]
            if t.type=='Valid':
                green=min(1,float(abs(t.seg.iend-t.seg.ibeg))/10)
                coul=r.rgb(1-green,1,1-green)
            xx=xcoord[Annotations[t.type]]
            #r.segments(xx,-t.beg,xx,-t.end,lwd=10,col=coul)
            if t.seg.len>1:
                r.polygon([xx-0.1,xx-0.1,xx+0.1,xx+0.1],[-t.beg,-t.end,-t.end,-t.beg],col=coul,border='NA')
            else:
                r.segments(xx-0.1,-t.beg,xx+0.1,-t.beg,col=coul)
        for b in self.bits:
            if (not reversed and not b.inv) or (reversed and b.inv) :
                coul='wheat'
            else:
                coul='grey'
            if len(b.mk)>1:
                r.polygon([0,0,1,1],[-b.bpbeg,-b.bpend,-b.bpend,-b.bpbeg],col=coul,border='NA')
            else:
                r.segments(0,-b.bpbeg,1,-b.bpbeg,col=coul)
        r.polygon([0,0,1,1],[ybeg,yend,yend,ybeg])
        #r.pdf(file=title[:-4]+'_mapcomp.pdf',width=5,height=-yend*0.2/1000000,pointsize=16)
        r.dev_off()


def RemoveMarkers(grplist):
    fathers=list(set([grp.father for grp in grplist]))
    newlist=[]
    for f in fathers:
        f.group=list(f.group)
        poubelle=[]
        for c in f.group:
            if c in grplist:
                poubelle.append(c)
        for merdasse in poubelle:
            merdasse.father=None
            f.group.remove(merdasse)
        f.group=tuple(f.group)
        if len(f.group)==0:
            newlist.append(f)
        elif len(f.group)==1:
            ## link granddad and grandson directly
            gdpere=f.father
            gdpere.group=list(gdpere.group)
            idx=gdpere.group.index(f)
            gdpere.group[idx]=f.group[0]
            gdpere.group=list(gdpere.group)
            f.group[0].father=gdpere
            ## unlink father
            f.father=None
            f.group=None
    if newlist==[]:
        return
    return RemoveMarkers(newlist)

def CheckIntervals(node,assembly):
    '''
    Check if markers in node form a common interval with the assembly
    This function also checks child groups.
    '''
    if node.type=='marker':
        try:
            node.mk_idx=[assembly.index(node.name)]
        except: 
            print node.code,node.name,'not in map'
            raise
        return node.mk_idx
    ## else : we are at an internal node
    node.mk_idx=[]
    for child in node.group:
        node.mk_idx+=CheckIntervals(child,assembly)
#         ## we don't have this information yet
#         child.father=node
    if (max(node.mk_idx)-min(node.mk_idx)+1)!=len(node.mk_idx):
        node.isinterval=False
        node.brokeby=[x for x in assembly[min(node.mk_idx):max(node.mk_idx)] if x not in node.show(with_marker_name=True)]
    else:
        node.isinterval=True
    return node.mk_idx

def FindBrokenGroups(node):
    if node.type=='marker':
        return []
    if not node.isinterval:
        count=[node]
    else:
        count=[]
    for child in node.group:
        count+=FindBrokenGroups(child)
    return count

def FindBreakingMarkers(nodes,markers):
    '''
    nodes: a list of broken nodes with a .brokeby attribute
    markers : a dictionary of markers {code: Marker instance}
    '''
    for n in nodes:
        mks=[x for x in markers.values() if x.name in n.brokeby]
        assert(len(mks)==len(n.brokeby))
        n.brokeby=mks
    return


def FindBreakingGroups(broken):
    #print 'FIND BREAKING GROUPS'
    ## consider sons before fathers <=> proceed from smallest to largest nodes
    broken=sorted(broken,cmp=lambda g1,g2: len(g1.show())-len(g2.show()))
    for grp in broken:
        grp.bkOrigin=[]
        ## 1. tries to find smallest group of origin of the breaking markers within the broken groups
        for grp2 in broken: ## we try to explain by the smallest groups first
            if grp is grp2:
                continue
            match=[mk for mk in grp.brokeby if mk.name in grp2.show(with_marker_name=True)]
            if match==[]:
                continue
            if len(match)==1:
                ## a single marker from grp2 breaks grp : find it
                # we add the father node to create connected components between groups != markers
                grp.bkOrigin.append(match[0])
            else:
                curgrp=grp2
                finalgrp=None
                while finalgrp==None:
                    got_milk=True
                    for child in curgrp.group:
                        mk_in_child=[mk for mk in match if mk.name in child.show(with_marker_name=True)]
                        if len(mk_in_child)==len(match):
                            curgrp=child
                            got_milk=False
                            break
                    if got_milk:
                        finalgrp=curgrp
                grp.bkOrigin.append(finalgrp)
            curgrp=grp
            while not curgrp.father.isinterval:
                for mk in match:
                    if mk in curgrp.father.brokeby:
                        curgrp.father.brokeby.remove(mk)
                if len(curgrp.father.brokeby)==0:
                    ## father breakage explained by sons entirely
                    #sys.stdout.write('getting rid of %s : '%node_ref(curgrp.father))
                    try:
                        broken.remove(curgrp.father)
                        #print ' Hasta La Vista Baby !'
                    except ValueError:
                        #print ' Been there, Done that'
                        pass
                curgrp=curgrp.father
            for mk in match:
                grp.brokeby.remove(mk)
        ## 2. Sometimes whole unbroken groups breaks a group: find them
        if len(grp.brokeby)>0:
            #print 'Still',len(grp.brokeby),'markers to find'
            # remove them from the fathers (we are going to explain them)
            curgrp=grp
            while not curgrp.father.isinterval:
                for mk in grp.brokeby:
                    if mk in grp.father.brokeby:
                        curgrp.father.brokeby.remove(mk)
                if len(curgrp.father.brokeby)==0:
                    #sys.stdout.write('getting rid of %s : '%node_ref(curgrp.father))
                    try:
                        broken.remove(curgrp.father)
                        #print ' Hasta La Vista Baby !'
                    except ValueError:
                        #print ' Oups, I did it again'
                        pass
                curgrp=curgrp.father
            # identifications of all father nodes of breaking markers
            bkfathers=[x.father for x in grp.brokeby]
            #print len(grp.brokeby),len(set(bkfathers))
            for fgrp in set(bkfathers):
                for mk in fgrp.group:
                    if mk in grp.brokeby:
                        grp.brokeby.remove(mk)
                        if bkfathers.count(fgrp)==1:
                            grp.bkOrigin.append(mk)
                            break
                if bkfathers.count(fgrp)>1:
                    grp.bkOrigin.append(fgrp)
    return broken
                 
            
            
                

def BreakGraph(broken,name='BKgaph.svg'):
    if has_yap:
        graph=yapgvb.Digraph('Graph of Broken Groups')
    ## utilisation de nx pour trouver les composantes connexes
    graph2=nx.Graph()
    nodes={}
    for grp in broken:
        if has_yap:
            if grp not in nodes.keys():
                nodes[grp]=graph.add_node(label=node_name(grp))
        graph2.add_node(grp)
        for grp2 in grp.bkOrigin:
            if grp2.type=='marker':
                bk=grp2.father
            else:
                bk=grp2
            if has_yap:
                if bk not in nodes.keys():
                    nodes[bk]=graph.add_node(label=node_name(bk))
                edge=nodes[grp]>>nodes[bk]
            graph2.add_edge(grp,bk)
    if has_yap:
        graph.layout(yapgvb.engines.dot)
        graph.render(name)
    return nx.connected_components(graph2)
    
        
def SolveTranslation(broken_node,breaker):
    broken_node.bkOrigin.remove(breaker)
    return broken_node

def ComponentGetMarkers(nodes):
    mkcodes=[]
    for node in nodes:
        mkcodes+=node.mk_idx
    mkcodes=list(set(mkcodes))
    mkcodes.sort()
    return mkcodes

def FindBits(markers,assembly,check_order=False):
    '''
    Get a list of markers (objects) and finds common intervals with assembly map
    '''
    ## make sure markers are sorted according to their position on the map (ie their codes)
    markers.sort(key=lambda a:a.code)
    if check_order:
        mkorder=sorted(markers,key=lambda x: assembly.index(x.name))
        ## we just want to check the ordering not the common interval problem = ignore gaps
        mkidx=[markers.index(m) for m in mkorder]
    else:
        mkidx=[assembly.index(m.name) for m in markers]
    L=[x-y for x,y in zip(mkidx[1:],mkidx[:-1])]
    breaks=[int(abs(x)>1) for x in L]
    bits=[]
    while True:
        try:
            bpidx=breaks.index(1)
        except ValueError:
            bits.append(tuple(markers))
            break
        b=markers[:bpidx+1]
        markers=markers[bpidx+1:]
        bits.append(tuple(b))
        breaks=breaks[bpidx+1:]
    return bits

