'''
Metamap module to build framework maps
'''

from carthagene.metamap import Groups
from carthagene.metamap.Report import node_ref
from math import log

class FrameworkMap:
    '''
    A Framework map class.
    
    A list of markers defining the Framework map
    A list of marker bins with their tags
    
    The framework map can be constructed when instanciated if a metamap is provided 
    at construction
    '''
    def __init__(self,metamap=None,probmin=1):
        self.probmin=probmin
        self.markers=[]
        self.bins=[]
        if metamap is not None:
            self.build(metamap,self.probmin)

    def build(self,metamap,probmin):    
        '''
        Build a framework map from the metamap.
        probmin: minimum probability for a node to be considered as certain
        returns:
        - markers for the framework map
        - bins of markers that cannot be included in the framework map
        '''
        if metamap.IsMkSeq():
            self.markers=[mk for mk in metamap.group]
            return
        CleanRoot(metamap)
        FindFWMarkers(metamap,probmin,self.markers,self.bins)

class FWbin:
    '''
    A marker bin class for framework maps

    A bin is a set of marker groups belonging to a node on the metamap which 
    cannot be ordered with confidence. One of the subnodes is chosen to 
    tag the bin.    
    '''
    def __init__(self,groups=None,tag=None):
        self.groups=groups
        self.tag=tag
    def markers(self):
        mks=[]
        for g in self.groups:
            mks+=g.show(with_marker_name=True)
        return mks

def GetChildrenOfHighProbability(node,probmin):
    if node.type=='marker':
        return [node]
    elif node.IsMkSeq() and (node.highP > probmin):
        return [node]
    else:
        rez=[]
        for grp in node.group:
            rez+=GetChildrenOfHighProbability(grp,probmin)
    return sorted(rez,key=lambda x:len(x),reverse=True)

def FindFWMarkers(node,probmin,markers,bins):
    if node.type=='marker':
        return
    if node.highP > probmin:
        if node.IsMkSeq():
            markers+=[mk for mk in node.group]
        else:
            for grp in node.group:
                FindFWMarkers(grp,probmin,markers,bins)
    else: ## uncertain node
        tag=GetChildrenOfHighProbability(node,probmin)[0]
        markers+=tag.markers()
        bins+=[FWbin(node.group,tag=tag)]
    ##print 'Worked on',node_ref(node),len(markers),len(bins)

def CleanRoot(metamap):
    '''
    Clean the metamap root from jumping markers
    Remove markers that are linked at the root of the metamap = unlinked markers
    '''
    children=metamap.group[:]
    orders={}
    for order in metamap.possible_order.items():
        orders[order[0]]=[list(order[0]),order[1]]
    pb=False
    metamap.group=list(metamap.group)
    for grp in children:
        if grp.type=='marker':
            pb=True
            print 'Removing',grp.code
            metamap.group.remove(grp)
            # remove the marker from the possible orders
            for pair in orders.values():
                pair[0].remove(grp.code)
    # 
    metamap.group=tuple(metamap.group)
    if pb:
        new_orders={}
        for pair in orders.values():
            try:
                new_orders[tuple(pair[0])]+=pair[1]
            except KeyError:
                new_orders[tuple(pair[0])]=pair[1]
        metamap.possible_order=new_orders
        metamap.highP=max(metamap.possible_order.values())
    return
