from __future__ import division

import Maps
import sys
from os.path import basename

class Data:
    """
    Generic class for carthagene input data
    """
    def __init__(self):
        self.type=""
        self.mknames=[]
        self.nmk=0
        self.name=""
    def __repr__(self):
        stdata="CG dataset: "+self.name+" - "
        stdata+="Type "+self.type+" - "
        stdata+=str(self.nmk)+" markers."
##        stdata+=" ".join([str(mk) for mk in self.mknames]) 
        return stdata
    def __str__(self):
        return "CG dataset: "+self.name
    
        
class RhData(Data):
    """
    Class for carthagene RH data input
    == Attributes ==
    type: rh
    rhvectors: dictionnary of RH vectors indexed by marker names
    """
    def __init__(self):
        Data.__init__(self)
        self.type="rh"
        self.rhvect={}
        self.nobs=0
    def ParseRhInput(self,f):
        self.name= basename(f.name)
        ligne=f.readline()
        buf=ligne.split()
        if (buf[0]=="data"):
            ## we have the data type line
            if (buf[2] != "radiated"):
                raise Exception("Bad File Format")
            ## get next line
            ligne=f.readline()
        buf=ligne.split()
        try:
            self.nobs=int(buf[0])
            self.nmk=int(buf[1])
        except:
            raise Exception("Bad File Format")
        data=f.readlines()
        for ligne in data:
            buf=ligne.split()
            self.rhvect[buf[0][1:]]=buf[1]
        self.mknames=self.rhvect.keys()
        self.mknames.sort()
        return None
    def write(self,fname=None,mk=None):
        if fname is None:
            f=sys.stdout
        else:
            f=open(fname+"_"+self.name,"w")
        if mk is None:
            markers=self.mknames
        else:
            markers=list(set(mk).intersection(self.mknames))
        markers.sort()
        print >> f, "data type radiated hybrid"
        print >> f, "%d %d 0 0" % (self.nobs,len(markers))
        for m in markers:
            print >> f, "*%s %s" % (m,self.rhvect[m])
        if f is not sys.stdout:
            f.close()
        return None
    def txret(self):
        txret={}
        for mk in self.mknames:
            ntot=self.nobs
            nh=0
            for hyb in self.rhvect[mk]:
                if hyb=='-':
                    ntot-=1
                elif hyb=='H':
                    nh+=1
            txret[mk]=nh/ntot
        return txret
            
class OrderData(Data):
    """
    Class for carthagene reference order data
    == Attributes ==
    type: order
    position: dictionnary of ref order position (chromosome, pos)
    indexed by marker names
    """
    def __init__(self):
        Data.__init__(self)
        self.type="order"
        self.position={}
        self.ord=None ## **ordered list of markers**
    def ParseOrdInput(self,f):
        self.name=basename(f.name)
        ligne=f.readline()
        buf=ligne.split()
        if (buf[0]=="data"):
            ## we have the data type line
            if (buf[2] != "order"):
                raise Exception("Bad File Format")
            ## get next line
            ligne=f.readline()
        buf=ligne.split()
        try:
            self.nmk=int(buf[0])
        except:
            raise Exception("Bad File Format")
        data=f.readlines()
        curchr=None
        self.ord=[]
        for ligne in data:
            buf=ligne.split()
            if (buf==[]):
                continue
            if (buf[0]!="*"):
                try:
                    curchr=int(buf[0])
                except:
                    raise Exception("Bad File Format")
            pos=int(buf[2])
            self.position[buf[1]]=(curchr,pos)
            self.ord.append(buf[1])
        self.ord=tuple(self.ord)
        self.mknames=self.position.keys()
        self.mknames.sort()
        return None
    def getmap(self,mk=None):
        if self.ord is None:
            if mk is None:
                markers=self.mknames
            else:
                markers=list(set(mk).intersection(self.mknames))
            markers=dict([(x,self.position[x]) for x in markers])
            ## sort it according to (chr,position) and keep only marker names
            markers=[x[0] for x in sorted(markers.items(),key=lambda (k,v): (v,k))]
        else:
            if mk is None:
                markers=self.ord
            else:
                subset=list(set(mk).intersection(self.ord))
                ## We want to keep the order right
                markers=[x for x in self.ord if x in subset]
        return(Maps.Map(markers))
    def write(self,fname=None,mk=None):
        """
        Output dataset in carthagene format
        """
        if fname is None:
            f=sys.stdout
        else:
            f=open(fname+"_"+self.name,"w")
        markers=self.getmap(mk)
        print >> f, "data type order"
        print >> f, "%d 0" % (len(markers.map))
        i=0
        curchr=None
        for m in markers.map:
            if (i==0):
                curchr=self.position[m][0]
                print >> f, "%d %s %s" % (curchr,m,self.position[m][1])
            elif (i==len(markers.map)-1):
                print >> f, "%d %s %s" % (curchr,m,self.position[m][1])
            else:
                mprev=markers.map[i-1]
                mnext=markers.map[i+1]
                if (self.position[mnext][0]!=self.position[m][0]):
                    print >> f, "%d %s %s" % (curchr,m,self.position[m][1])
                elif (self.position[mprev][0]!=self.position[m][0]):
                    curchr=self.position[m][0]
                    print >> f, "%d %s %s" % (curchr,m,self.position[m][1])
                else:
                    print >> f, "* %s %s" % (m,self.position[m][1])
            i+=1
        if f is not sys.stdout:
            f.close()
        return None

class MergedData(Data):
    """
    Class for merged datasets
    == Attributes ==
    type: merged
    datasets: the datasets that are merged
    """
    def __init__(self,datasets):
        if (len(datasets) < 2):
            print "datasets arguments must be of size >= 2"
            return
        Data.__init__(self)
        self.type="merged"
        self.datasets=list(datasets)
        ## get list of markers common to all datasets
        mkset=set(datasets[0].mknames)
        for d in datasets[1:]:
            mkset=mkset.intersection(d.mknames)
        self.mknames=list(mkset)
        self.mknames.sort()
        self.nmk=len(self.mknames)
        self.name="Merged("+",".join([x.name for x in datasets])+")"
    def write(self,fname=None,mk=None):
        if mk is None:
            markers=self.mknames
        else:
            markers=list(set(mk).intersection(self.mknames))
        i=0
        for d in self.datasets:
            print "Calling Write for ", d
            if d.type is "merged":
                if fname is None:
                    d.write(fname,markers)
                else:
                    d.write(fname+"2",markers)
            else:
                if fname is None:
                    d.write(fname,markers)
                else:
                    if (d.name==""):
                        d.write(fname+"_"+str(i)+"_"+d.type,markers)
                        i+=1
                    else:
                        d.write(fname,markers)
                        
        return None

        
def ParseInput(fname=None):
    """
    Parse carthagene input file.
    arguments:
    fname: file name to read from (default = sys.stdin)
    """
    if (fname==None):
        f=sys.stdin
    else:
        f=open(fname,"r")
    buf_type=f.readline()
    buf_type=buf_type.split()
    if (buf_type[2]=="order"):
        print "Order Data Type"
        data=OrderData()
        data.ParseOrdInput(f)
    elif (buf_type[2]=="radiated"):
        print "RH Data Type"
        data=RhData()
        data.ParseRhInput(f)
    f.close()
    return data

def ParseLkhOutput(stream):
    """
    Parse Output from the lkh command from stream f

    Arguments:
    stream: stream to parse output from

    Return Value:
    LKH map instance
    """
    lkh=None
    while lkh==None:
        ligne=stream.next()
        buf=ligne.split()
        if buf==[]:
            continue
        if buf[0]=="1":
            lkh=Maps.Map(buf[2:])
    return lkh
            

def ParseCgLog(fname,proceed=False):
    """
    Parse carthagene session output (as written by cgout)

    Arguments:
    fname: file name to read from (no default)
    proceed: import data from input files (default = False)
    
    Return Value:
    - list of input files
    - LKH map instance
    - merged Datasets
    """
    f=open(fname,"r")
    inputfiles=[]
    lkh=None
    mergedDsets=[]
    datasets=[]
    for ligne in f:
        buf=ligne.split()
        if buf==[]:
            continue
        if buf[0]=="CG_log>":
            ## read a command
            if buf[1].find("dsload")!=-1:
                inputfiles.append(buf[2])
                continue
            elif buf[1].find("lkh")!=-1:
                ## parse lkh output
                lkh=ParseLkhOutput(f)
                continue
            elif buf[1].find("dsmergor")!=-1:
                mergedDsets.append( (int(buf[2])-1,int(buf[3])-1) )
            else:
                print "Command not parsed: ",buf[1]
    if (proceed):
        ## Get Data
        for dset in inputfiles:
            datasets.append(ParseInput(dset))
            for dset in mergedDsets:
                tomerge=[datasets[x] for x in dset]
                datasets.append(MergedData([datasets[x] for x in dset]))
    
    return(inputfiles,lkh,mergedDsets,datasets)
                
                    
