import Maps
import Groups
import CommonIntervals
import copy
from math import log


####################################################################################
#                               Methods                              
####################################################################################


#-----------------------------------------------------------------------------------
#                             CREATE MAPS WITH A SELECTION OF GROUPS                 
#-----------------------------------------------------------------------------------


def ReduceMaps(partition,maps):
	'''
	INPUT : partition: a dictionary with the code of groups as key and the group object as values
		maps= a list of maps, they contain as list the codes of the children of each group
	OUTPUT : reduced_maps : a list with the reduce maps
	''' 
	reduced_maps=[]
	
	for my_map in maps:
		index_vs_group={}
		for a_group in partition.values() :
			group_index=[my_map.index(x) for x in a_group.show()]
			if group_index!=[] :
				index_vs_group[min(group_index)]=a_group.code
		new_map=[index_vs_group[j] for j in  sorted(index_vs_group.keys())]
		reduced_maps.append(Maps.Map(new_map,my_map.prob))
	return reduced_maps


#--------------------------------------------------------------------------------------
#                             SELECT MAPS                 
#--------------------------------------------------------------------------------------
def MapsSelection(maps,groups):
	''' 
	This fuction separated maps where all the children of each given 
	group form a Common Interval from those that don't
	INPUT : maps : all the maps that contain the code of the children of the groups
		groups  : the groups you are looking for in the maps
	OUTPUT : selected_maps,problematic_maps encoded with the group codes
	'''
	selected_maps=[]
	problematic_maps=[]
	for my_map in maps:
		order=list(my_map.map)
		is_problematic=False
		for i in groups :
			index_my_group=[order.index(x) for x in groups[i].show_components()]
			if max(index_my_group)-min(index_my_group)+1 != len(groups[i]) :
				is_problematic=True
				break
			else :
				order[min(index_my_group)]=min(groups[i].show_components())
				for i in range(1,len(groups[i])) :
					del order[min(index_my_group)+1]
		if is_problematic :
			problematic_maps.append(Maps.Map(order, my_map.prob))
		else :
			selected_maps.append(my_map)
	return selected_maps,problematic_maps


#--------------------------------------------------------------------------------------
#                                       BUILD THE MATRIX DISTANCE                
#--------------------------------------------------------------------------------------


def BuildDistanceMatrix (distribution):
	'''
	Build the matrix of distance betwen each marker.
	input : distribution : a list of maps
	output : A (sparse matrix) M[i][j] = (pl,pr) pl = P(j is on the left of i)
	'''
	Mdist={}
	for m in distribution:
		for left,right in zip(m[:-1],m[1:]):
			if left not in Mdist.keys():
				Mdist[left]={}
			if right not in Mdist.keys():
				Mdist[right]={}
			try:
				Mdist[left][right][1]+=m.prob
			except KeyError:
				Mdist[left][right]=[0,m.prob]
			try:
				Mdist[right][left][0]+=m.prob
			except KeyError:
				Mdist[right][left]=[m.prob,0]
	return Mdist


#--------------------------------------------------------------------------------------
#                        CREATE THE MATRIX DISTANCE FOR A SET OF GROUPS                
#--------------------------------------------------------------------------------------


def GetNewMatrix(groups,old_matrix) :
	'''
	Computes the Distance Matrix between groups of markers (in groups)
	The distance is computed as the SUM(m1 in g1)(m2 in g2) d(m1,m2)
	d(m1,m2) is old_matrix
	'''
	new_matrix={}
	for i in range(len(groups)) :
		for j in range(i+1,len(groups)) :
			g1=groups.values()[i]
			g2=groups.values()[j]
			d=[0.0,0.0]
			for m1 in g1.show() :
				for m2 in g2.show() :
					if m2 in old_matrix[m1] :
						d=[d[k]+old_matrix[m1][m2][k] for k in range(2)]
			if d!=[0.0,0.0]:
				try :
					new_matrix[g1.code][g2.code]=d
				except KeyError:
					new_matrix[g1.code]={g2.code:d}
				try :
					new_matrix[g2.code][g1.code]=d
				except KeyError:
					new_matrix[g2.code]={g1.code:d}
	return new_matrix	


#--------------------------------------------------------------------------------------
#                                       BUILD SEQUENCES                
#--------------------------------------------------------------------------------------



#We will build sequence to do that we build the connex composants of a graph. This graph is define by the distance matrix with a step of 1-precision

def  BuildSequences(distance_matrix,partition,precision=10e-8):
	'''
	build a new partion by building sequences
	If the precision is 10e-10  the new groups are sequences if the precision for exemple is 0.01 the father groups 
	are probable sequence but in some maps these groups are broken and in more maps these groups are not sequences
	INPUT : distance_matrix : the distance betwen the children groups
	INOUT :	partition : the partition of the maps
	'''
	#Build the list of the linked couple components 
	edges=BuildGraph(distance_matrix,precision)
	
	## Find groups to merge
	## new_groups: init at actual groups
	new_groups=[]
	for k in partition.keys():
		new_groups.append([k])
	#We will join the fathers that have linked children, at the end we will have the father groups we are looking for
	for my_edge in edges:
		group1=FindGroup(new_groups,my_edge[0])
		group2=FindGroup(new_groups,my_edge[1])
		if group1 != group2:
			## Attach groups by the edge
			if group1.index(my_edge[0])==0 :
				group1.reverse()
			if group2.index(my_edge[1])==(len(group2)-1):
				group2.reverse()
			new_groups.remove(group1)
			new_groups.remove(group2)
			new_groups.append(group1+group2)
	## Merge Groups:
	for codes in new_groups:
		if len(codes)==1:
			continue
		tomerge=[partition[i] for i in codes]
		new_grp=Groups.Sequence(tomerge)
		for g in new_grp.group:
			g.father=new_grp
		for i in codes:
			del partition[i]
		partition[new_grp.code]=new_grp
	
def BuildGraph(distance_matrix,precision) :
	'''
	permit to find the couple of components that are connected : these couple are composed by the
	codes of two children groups that are linked with probability 1-precision
	INPUT :distance_matrix : the distance matrix beywen the vertices
		the precision to build the connection
	OUTPUT : edges : a list with all the couples
	'''
	edges=[]
	for i in distance_matrix:
		for n in distance_matrix[i]:
			intermediate_list=[(i,n) for n in distance_matrix[i] if abs(sum(distance_matrix[i][n])-1)<precision]
		for j in intermediate_list :
			if (j[0],j[1]) not in edges and (j[1],j[0]) not in edges:
				edges.append(j)	
	return edges

def FindGroup(groupes,my_group):
	'''
	Find the group in groupes containinf my_group
	INPUT : groupes : list of groups
		my_group : a group
	'''
	for grp in groupes:
		if  my_group in grp:
			return grp

#--------------------------------------------------------------------------------------
#            FIND ALL THE POSSIBLE ORDER FOR EACH PART OF A PARTITION                  
#--------------------------------------------------------------------------------------


def GetGroupsOrders(node,maps):
	if node.type=='marker':
		return
	for myMap in maps:
		order=[]
		for grp in node.group:
			order.append((myMap.index(grp.code),grp.code))
		# sort the groups according to their index on the map and keep only the ordering
		order=tuple([x[1] for x in sorted(order,key=lambda x:x[0])])
		try:
			node.possible_order[order]+=myMap.prob
		except KeyError:
			node.possible_order[order]=myMap.prob
	try:
		node.highP=max(node.possible_order.values())
	except:
		print p
	return
#--------------------------------------------------------------------------------------
#                                         DEL MARKER IN A TREE                      
#--------------------------------------------------------------------------------------


# def ReduceTree(markers_to_delete,all_markers) :
# 	'''
# 	this function can be use only and only if the marker to delete have a father. So is better to use it when the tree is finish
# 	reduction of the tree by remove some link between a father and its marker child
# 	INPUT : a list that contain all the codes of the marker to delete
# 		all_markers : a dictionary that associate each code of marker to a marker
# 	INOUT : A modifictation of the tree	
# 	'''
# 	for my_marker_code in markers_to_delete :
# 		my_father_group=[ x for x in all_markers[my_marker_code].father.group if x.code != my_marker_code]
# 		all_markers[my_marker_code].father.group=tuple(my_father_group)


def ReDoTree(deleted_markers,all_markers) :
	'''
	Recreate a link betwen a father and a deleted marker
	INPUT : a list that contain all the codes of the marker to delete
		all_markers : a dictionary that associate each code of marker to a marker
	INOUT : A modifictation of the tree
	'''
	for my_marker_code in deleted_markers :
		all_markers[my_marker_code].father.group=tuple(list(all_markers[my_marker_code].father.group)+[all_markers[my_marker_code]])

#--------------------------------------------------------------------------------------
#                             Find Order      
#--------------------------------------------------------------------------------------

def FindOrders(node, maps):
	'''
	Compute the possible order of the children for each node
	INOUT : my partition
	INPUT the maps
	'''
# 	n=0
# 	depth_n_partition=my_partition
# 	while len(depth_n_partition)!=len(maps[0]) :
# 		n=n+1
# 		GetGroupsOrders(depth_n_partition,maps)
# 		depth_n_partition=my_partition[0].subgroups(n)
	
	GetGroupsOrders(node,maps)
	if node.type=='marker':
		return
	for grp in node.group:
		FindOrders(grp,maps)

#--------------------------------------------------------------------------------------
#                             Genome assembly validation
#--------------------------------------------------------------------------------------

#usefull function
def FindParts(the_values):
	'''
	Find the suite of consecutive numbers in a list of numbers
	INPUT : the list with the number (no sorted)
	OUTPUT : a list of list of interval of consecutive number of the input list
	'''
	my_parts=[]
	x=0
	a_part=[]
	#We sort the values
	sorted_values=sorted(the_values)
	#We are looking for interval wher the number are consecutive
	while x < len(sorted_values) :
		if len(a_part)==0 :
			a_part.append(sorted_values[x])
		elif  a_part[-1]+1==sorted_values[x]:
			a_part.append(sorted_values[x])
		else :
			my_parts.append(a_part)
			a_part=[sorted_values[x]]
		x=x+1
	my_parts.append(a_part)
	return my_parts


	

def FindNoise (my_partition,assembly,nb_markers=1):
	'''
	In : my_partition : my tree <=> my partition deph 0
	    the assembly
	    nb_markers : size ofthe part that are conciderate like noise
	Out : the noise with the marker codes
	'''
	#Algorithm
	#Build the partition for each depth of the tree, begin by the upper noddes
	#For each deph:
	#	For each nodle :
	#		if it is broken in assembly
	#			calculate all the parts
	#			if some part have a size of one marker and are in the extremity,
	#			add them to the noise 
	
	n=0
	#the noise all the partitions
	all_noise=[]
	#partion deph 0
	depth_n_partition=my_partition
	while len(depth_n_partition)!=len(assembly):
		n+=1
		depth_n_partition=my_partition[0].subgroups(n)
		#the noise for one partition
		noise=[]
		for node in [x for x in depth_n_partition.values() if len(x.show())>1]:
			#computing of the rate :
			#(index of the last child of the nodle in the assembly- the index of the firts +1)/ number of marker for the nodle
			#if equal 1 the group is not broken, else it is broken
			
			index_in_assembly=[assembly.show().index(x) for x in node.show()]
			min_index=min(index_in_assembly)
			max_index=max(index_in_assembly)
			rate=float(len(node.show())) / float((max_index-min_index+1))	
			
			#When a node is broken Analyse of groups of children in assembly
			if rate !=1 :
				pieces_part=FindParts(index_in_assembly)
				pieces_part=sorted(pieces_part,key=lambda x:x[0])
				modif=True
				#all the single markers that are in the extremity are deleted, for exemple if we have a node cut like this :
				#1 marker (..) 1 marker (..)  5markers (..) 1 markers(..)  4 markers (..) 1 markers 
				# the part the  will not put in the noise is :  5markers(..) 1 markers (..)4 markers
				while modif==True :
					if len(pieces_part)!=0 :
						length_pieces=[len(x) for x in pieces_part]
						if (length_pieces[0]==nb_markers or length_pieces[-1]==nb_markers) :
							if length_pieces[0]==nb_markers:
								for j in assembly.show()[min(pieces_part[0]):max(pieces_part[0])+1]:
									if j not in noise:
										noise.append(j)
								del pieces_part[0]
							if length_pieces[-1]==nb_markers and not len(pieces_part)==0:
								for j in assembly.show()[min(pieces_part[-1]):max(pieces_part[-1])+1]:
									if j not in noise:
										noise.append(j)
								del pieces_part[-1]
						else :
							modif=False
					else :
						modif=False
		#update of the total noise
		if len(noise)!=0:
			for x in noise :
				if x not in all_noise :
					all_noise.append(x)
	return sorted(all_noise)








def NoiseAnalyse(reduce_assembly,my_partition,my_out_put_file,markers,positions,noise):
	'''
	Write a file with the noise analyse
	input : reduce_assembly : the assembly without the noise
		my_partition : the tree updated (witjhout the noised markers)
		markers : the dictionary that contain all the markers (marker code -> marker)
		my_out_put_file : the name of the file wher will be written the analyse
		positions : the dictionary taht conatin the position of each marker in the assembly (marker name -> position in asssembly)
		noise : a list of the markers that seems to be noise
	output : tje file with the analyse
	'''
	my_file=open(my_out_put_file,'w')
	print>>my_file,'code		name		from * 	      to * 	expected position'
	for marker in sorted(noise) :										
		index_group_in_assembly=[reduce_assembly.show().index(x) for x in markers[marker].father.show()]
		#We look the farther node of each marker:
		# if the nodle is empty we say taht is impossible to analyse
		#else : 
		#	if  this node is again broken we can say nothing for the hypothetic position of the conciderate marker
		# 	if the position of the considerate marker is in the middle of the father group : we can say nothing about the hypothetic position 
		# 	else we can give the gap between the marker an its father group
			
		if index_group_in_assembly!=[]:
			marker_position=positions[markers[marker].name][1]
			min_index=min(index_group_in_assembly)
			max_index=max(index_group_in_assembly)
			rate=float(len(markers[marker].father.show())) / float((max_index-min_index+1))	
			if rate==1 :
				max_gap_position=abs(positions[reduce_assembly.group[max_index].name][1]-marker_position)
				min_gap_position=abs(positions[reduce_assembly.group[min_index].name][1]-marker_position)
			elif  marker_position<positions[reduce_assembly.group[max_index].name][1] and marker_position>positions[reduce_assembly.group[min_index].name][1]:
				max_gap_position='in the middle'
				min_gap_position=''
			else :
				max_gap_position='-'
				min_gap_position='-'
		else :
			max_gap_position='deleted father group'
			min_gap_position=''
		print>>my_file,marker,'		', markers[marker].name ,'		',min([min_gap_position,max_gap_position]),'		',max([min_gap_position,max_gap_position])
	print>> my_file,'----------------------------------------------------'
	my_file.write('\n')
	print>>my_file,'noise :'
	print>>my_file,sorted(noise)
	print>>my_file,'length :' , len(noise)
	my_file.close()	




def Structure_Analyse(assembly,my_partition,my_out_put_file1,my_out_put_file2,markers,positions):
	'''
	INPUT : the assembly 
		my partition :the tree  (ther is the same number of markers in the tree and in the assembly)
		my_out_put_file1 : name of the file tht contain the anlayse of structure
		my_out_put_file2 : name of the file tht contain the anlayse of internal order of each group
		markers : the dictionary that contain all the markers (marker code -> marker)
		positions : the dictionary taht conatin the position of each marker in the assembly (marker name -> position in asssembly)
	OUTPUT : two file one with the structure analyse , one with the internal ordre of each group analyse
	'''
	my_file1=open(my_out_put_file1,'w')
	my_file2=open(my_out_put_file2,'w')
	print>>my_file1,'Analyse of the group structure '
	print>>my_file2, 'Analyse of the internal order of the groups'
	n=0
	depth_n_partition=my_partition
	
	while len(depth_n_partition)!=len(assembly):
		for my_part in [x for x in depth_n_partition.values() if len(x.show())>1]:
			
			#Analyse of each nodle of the deph n partition :
			# if it is broken we can't propone an order for its children 
			child_codes={}
			code_for_order={}
			my_file2.write('\n')
			index_in_assembly=[assembly.show().index(x) for x in my_part.show()]
			min_index_part=min(index_in_assembly)
			max_index_part=max(index_in_assembly)
			rate=float(len(my_part.show())) / float((max_index_part-min_index_part+1))
			
			if rate==1:
				variable='['+str(positions[assembly.group[min_index_part].name][1])+','+str(positions[assembly.group[max_index_part].name][1])+']'
				father_is_broken=False
			else:
				father_is_broken=True	
				variable='BROKEN'
			
			
			#Writing of the name of the analyse node
			#We don't write a nodle when its children are only markers
			if len([x.code for x in my_part.group if len(x)==1])!=len(my_part.show()):
				my_file1.write('\n')
				print>>my_file1,'##############',str( n)+'-'+str(my_part.code),'   #',len(my_part.show())
			print>>my_file2,'---------------------------------------------------------'
			print>>my_file2,str( n)+'-'+str(my_part.code),'		#',len(my_part.show()),'	', variable
			print>>my_file2,'---------------------------------------------------------'

			
			
			####################################################################################################
			#STRUCTURE ANALYSE 
			####################################################################################################
			
			not_broken_child=True
			all_pieces=[]
			
			
			#Position of the children of size more than 1
			for child in [x for x in my_part.group if len(x.show())>1]:	
				child_code=str( n+1)+'-'+str(child.code)
				child_codes[child.code]=child_code
				index_in_assembly=[assembly.show().index(x) for x in child.show()]
				min_index=min(index_in_assembly)
				max_index=max(index_in_assembly)
				rate=float(len(child.show())) / float((max_index-min_index+1))	
				if rate==1 :
					all_pieces.append((child_code,[positions[assembly.group[min_index].name][1],positions[assembly.group[max_index].name][1]],len(index_in_assembly)))
				else :
					
					not_broken_child=False
					pieces_part=FindParts(index_in_assembly)
					mins=sorted([min(x) for x in pieces_part])
					maxs=sorted([max(x) for x in pieces_part])
					for i in range(len(mins)):
						all_pieces.append(('*'+child_code,[positions[assembly.group[mins[i]].name][1],positions[assembly.group[maxs[i]].name][1]],maxs[i]-mins[i]+1))
					all_pieces=sorted(all_pieces,key=lambda x : x[1][0])
			
			
			#Position  of the children of size 1
			markers_position=[]
			for child in [x for x in my_part.group if len(x.show())==1]:
				child_codes[child.code]=child.show()[0]
				markers_position.append(positions[markers[child.show()[0]].name][1])
			
			
			#Writting of the position of all the child in the file
			if markers_position!=[] and len([x.code for x in my_part.group if len(x)==1])!=len(my_part.show()):
				my_file1.write('markers')
				my_file1.write('	')
				for x in markers_position:
					my_file1.write(repr([x]))
					my_file1.write('	')
				my_file1.write('#'+str(len(markers_position)))
				my_file1.write('\n')
			for x in all_pieces :
				print>>my_file1,x[0],'  ',x[1],'		#',x[2]
			
			
			####################################################################################################
			#INTERNAL ORDER ANALYSE 
			####################################################################################################

			if not_broken_child and not father_is_broken :
				
				#We have to update the possible order (eventualy remove the deletd markers and add identify the same orders)
				#at the end we have a list of map with the possible orders, ordonned thanks to their probability : the more probable  is first
				all_orders=my_part.possible_order
				for x in all_orders :
					x.map=tuple([child_codes[j] for j in x.map if j in child_codes.keys()])
				reduce_all_orders=[]
				while all_orders!=[]:
					similar_orders=[q for q in all_orders if q==all_orders[0]]
					reduce_all_orders.append(Maps.Map(list(all_orders[0].map),sum([x.prob for x in similar_orders])))
					all_orders=[x for x in all_orders if x not in similar_orders]
				all_orders=reduce_all_orders
				all_orders=sorted(all_orders,key=lambda i:i.prob)
				all_orders.reverse()
				
				
				#Compute the part of the assembly that correspond to the conciderate node
				assembly_map=[Maps.Map(assembly.show())]
				group_partition={}
				for x in my_part.group :
					group_partition[x.code]=x
				[reduce_assembly]=ReduceMaps(group_partition,assembly_map)
				my_assembly_order=[child_codes[x] for x in reduce_assembly.map if x in child_codes.keys()]

				
				# If the order of the assembly is lso in the distribution we give its probability and the alternative order with its probability else we say that the order of the assembly doesn't exist in the distribution and we give to alternative ordet with probability
				assembly_vs_orders=[ an_order  for an_order in all_orders if list(an_order.map) ==my_assembly_order]
				if len (assembly_vs_orders)==1:
					the_order=assembly_vs_orders[0]
					print>>my_file2,'assembly_order : ',my_assembly_order,'  prob=',"%.6f" %the_order.prob
					if the_order.prob<0.95:
						all_orders.remove(the_order)
						alternative_order=list(all_orders[0].map)
						print>>my_file2,'alternative order',alternative_order,'	prob=',"%.6f" %all_orders[0].prob
				else:
					print>>my_file2,'assembly_order : ', my_assembly_order,'  prob=NONE'
					if len(all_orders)>1 :
						print>>my_file2,'alternative orders'
						for an_order in all_orders[0:2]:
							alternative_order=list(an_order.map)
							print>>my_file2,alternative_order,'	prob=',"%.6f" %an_order.prob
					else :
						alternative_order=list(all_orders[0].map)
						print>>my_file2,'alternative order',alternative_order,'	prob=',"%.6f" %all_orders[0].prob
			if not not_broken_child :
				print>>my_file2,'BROKEN COMPONENTS'
		
		n+=1
		depth_n_partition=my_partition[0].subgroups(n)
	my_file1.close()
	my_file2.close()	
